include "c:\windows\\bureau\\Huff.ml";;
include "c:\windows\\bureau\\Zigzag.ml";;
(*                                                   U T I L I T A I R E S *)

(* Renvoie l'entier le plus proche de a/b *)
let rec div_int a b = match b with
  |1 -> a
  |2 -> if a>0 then (a+1) lsr 1 else -((1-a) lsr 1)
  |_ -> div_int (a/(b lsr 1)) 2;;

(* Renvoie l'entier le plus proche de a/2^n , a lsr n divise a supprime les n premiers chiffres 
      de l'criture en base 2 de a (~ divise a par 2^n) *)
let rec div_puiss2 a n = match n with
  | _ when a<0 -> -(div_puiss2 (-a) n)
  | 0 -> a
  | 1 -> (a+1) lsr 1
  | _ -> div_puiss2 (a lsr (n-1)) 1;;
  
(* Calcule a puissance n o a et n sont des entiers et n est positif *)
let rec puissance a n = match n with
  | 0 -> 1
  | 1 -> a
  | _ -> a*(puissance a (n-1));;

(* Renvoie dans un vecteur les cos(i*Pi/16) o i appartient  [|0,45|] *)
let cosinus =
  let c = make_vect 46 0 and Pi = 4.*.(atan 1.) in
    for i = 0 to 45 do
     c.(i) <- div_int (int_of_float (1024.*.cos (Pi*.(float_of_int i)/.16.))) 2
    done;
c;;

(* Effectue un "changement de base" (Y,U,V) -> (R,V,B) *)
let yuv_to_rvb (y,u,v) =
  let y2 = 128+y in
    (y2+(div_int (14*u) 10),y2-(div_int (344*v+714*u) 1000),y2+(div_int (177*v) 100)) ;;

(* Effectue un "changement de base" (R,V,B) -> (Y,U,V) *)    
let rvb_to_yuv (r,v,b) =
 ((div_int (299*r+587*v+114*b) 1000) - 128,
 (div_int (500*r-419*v-81*b) 1000) , (div_int (-168*r-331*v+500*b) 1000)) ;;

(* Transpose une matrice 8 * 8 *)  
let transpos m =
  let n = make_matrix 8 8 0 in
  for i = 0 to 7 do
    for j = 0 to 7 do
      n.(i).(j) <- m.(j).(i)
    done;
  done;
n;;

let align n = 
  if n<0 then 1 else
    if n>255 then 254 else n ;;

    
(*                                               L E  P R O G R A M M E  D C T *)

(* Applique la DCT sur la i colonne ou ligne de m en stockant les variables temporelles dans
   temp1 et temp2 *)
let formule_dct col_ou_ligne m i temp1 temp2 =
  if col_ou_ligne then
  ( for x = 0 to 3 do
      temp1.(x) <- m.(i).(x) + m.(i).(7-x);
      temp2.(x) <- m.(i).(x) - m.(i).(7-x);
    done ; )       else
  ( for x = 0 to 3 do
      temp1.(x) <- m.(x).(i) + m.(7-x).(i);
      temp2.(x) <- m.(x).(i) - m.(7-x).(i);
    done ;);
  temp1.(4) <- temp1.(0)+temp1.(3);
  temp1.(0) <- temp1.(0)-temp1.(3);
  temp1.(5) <- temp1.(1)+temp1.(2);
  temp1.(1) <- temp1.(1)-temp1.(2);
  if col_ou_ligne then
  ( m.(i).(0) <- div_puiss2 ((temp1.(4)+temp1.(5))*cosinus.(4)) 1 ;
    m.(i).(2) <- div_puiss2 (temp1.(0)*cosinus.(2)+temp1.(1)*cosinus.(6)) 1   ;
    m.(i).(4) <- div_puiss2 ((temp1.(4)-temp1.(5))*cosinus.(4)) 1  ;
    m.(i).(6) <- div_puiss2 (temp1.(0)*cosinus.(6)-temp1.(1)*cosinus.(2)) 1 ;
    m.(i).(1) <- div_puiss2 (temp2.(0)*cosinus.(1)+temp2.(1)*cosinus.(3)+temp2.(2)*cosinus.(5)+temp2.(3)*cosinus.(7)) 1;
    m.(i).(3) <- div_puiss2 (temp2.(0)*cosinus.(3)-temp2.(1)*cosinus.(7)-temp2.(2)*cosinus.(1)-temp2.(3)*cosinus.(5)) 1;
    m.(i).(5) <- div_puiss2 (temp2.(0)*cosinus.(5)-temp2.(1)*cosinus.(1)+temp2.(2)*cosinus.(7)+temp2.(3)*cosinus.(3)) 1;
    m.(i).(7) <- div_puiss2 (temp2.(0)*cosinus.(7)-temp2.(1)*cosinus.(5)+temp2.(2)*cosinus.(3)-temp2.(3)*cosinus.(1)) 1; )
      else
  ( m.(0).(i) <- div_puiss2 ((temp1.(4)+temp1.(5))*cosinus.(4)) 19 ;
    m.(2).(i) <- div_puiss2 (temp1.(0)*cosinus.(2)+temp1.(1)*cosinus.(6)) 19 ;
    m.(4).(i) <- div_puiss2 ((temp1.(4)-temp1.(5))*cosinus.(4)) 19 ;
    m.(6).(i) <- div_puiss2 (temp1.(0)*cosinus.(6)-temp1.(1)*cosinus.(2)) 19 ;
    m.(1).(i) <- div_puiss2 (temp2.(0)*cosinus.(1)+temp2.(1)*cosinus.(3)+temp2.(2)*cosinus.(5)+temp2.(3)*cosinus.(7)) 19 ;
    m.(3).(i) <- div_puiss2 (temp2.(0)*cosinus.(3)-temp2.(1)*cosinus.(7)-temp2.(2)*cosinus.(1)-temp2.(3)*cosinus.(5)) 19 ;
    m.(5).(i) <- div_puiss2 (temp2.(0)*cosinus.(5)-temp2.(1)*cosinus.(1)+temp2.(2)*cosinus.(7)+temp2.(3)*cosinus.(3)) 19 ;
    m.(7).(i) <- div_puiss2 (temp2.(0)*cosinus.(7)-temp2.(1)*cosinus.(5)+temp2.(2)*cosinus.(3)-temp2.(3)*cosinus.(1)) 19 ; )
;;

(* Applique la DCT  une matrice 8*8 *)
let dct m =
  let temp1=make_vect 6 0 and temp2 = make_vect 4 0 in
  for i=0 to 7 do
    formule_dct true m i temp1 temp2
  done;
  for j=0 to 7 do
    formule_dct false m j temp1 temp2
  done;
m;;

(*                                 L E  P R O G R A M M E  I D C T  *)

(* Effectue une IDCT sur un vecteur *)
let idct_1d t =
  let a = make_vect 4 0 and b=make_vect 4 0 and c=cosinus in
  b.(0) <- (t.(0)+t.(4))*c.(4); b.(1) <- (t.(0)-t.(4))*c.(4);
  b.(2) <- t.(2)*c.(2)+t.(6)*c.(6); b.(3) <- t.(2)*c.(6)-t.(6)*c.(2);
  a.(0) <- b.(0)+b.(2) ; a.(1) <- b.(1)+b.(3) ; a.(2) <- b.(1)-b.(3) ; a.(3) <- b.(0)-b.(2);
  b.(0) <- t.(1)*c.(1)+t.(3)*c.(3)+t.(5)*c.(5)+t.(7)*c.(7);
  b.(1) <- t.(1)*c.(3)-t.(3)*c.(7)-t.(5)*c.(1)-t.(7)*c.(5);
  b.(2) <- t.(1)*c.(5)-t.(3)*c.(1)+t.(5)*c.(7)+t.(7)*c.(3);
  b.(3) <- t.(1)*c.(7)-t.(3)*c.(5)+t.(5)*c.(3)-t.(7)*c.(1);
[|div_int (a.(0)+b.(0)) 2;div_int (a.(1)+b.(1)) 2;div_int (a.(2)+b.(2)) 2;div_int (a.(3)+b.(3)) 2;
    div_int (a.(3)-b.(3)) 2;div_int (a.(2)-b.(2)) 2;div_int (a.(1)-b.(1)) 2;div_int (a.(0)-b.(0)) 2|];;

(* Calcule une IDCT sur chacune des lignes de la matrice, et renvoie sa transpose *)
let idct1d_on_matrix m =
 transpos [|idct_1d m.(0);idct_1d m.(1);idct_1d m.(2);idct_1d m.(3);idct_1d m.(4);
              idct_1d m.(5);idct_1d m.(6);idct_1d m.(7)|];;

(* Renvoie la DCT INVERSE d'une matrice *)
let idct m =
  let n = idct1d_on_matrix (idct1d_on_matrix m) in
  for i=0 to 7 do
    for j=0 to 7 do
      n.(i).(j) <- div_int n.(i).(j) 262144
    done;
  done;
n;;

(*                                     Q U A N T I Z A T I O N *)

(* "Quantize" une matrice 8*8 par une matrice q *)
let quantiz m q =
  for i = 0 to 7 do
    for j = 0 to 7 do
      m.(i).(j) <- div_int (2*m.(i).(j)) (2*q.(i).(j))
    done;
  done;
m;;

(* "Dquantize" une matrice 8*8 par une matrice q *)
let dequantiz m q =
  for i = 0 to 7 do
    for j = 0 to 7 do
      m.(i).(j) <- (m.(i).(j))*(q.(i).(j))
    done;
  done;
m;;

(* Cre une matrice de quantization standard avec la formule (1+T*(1+i+j)^Q) *)
let kant t qualite =
  let q = make_matrix 8 8 0 in
  for i = 0 to 7 do
    for j = 0 to 7 do
      q.(i).(j) <- (1+ (puissance (1+t*(i+j)) qualite))
    done;
  done;
q;;

let test m =
  let k= kant 1 1 in
  dequantiz ( quantiz m k) k;;

(*                             RELATION PRG <-> DISQUE *)

(* Cre 3 matrices de matrices 8*8 indpendantes *)
let make_3matofmat larg haut =
    let j_max=(larg-1)/16+1 and i_max=(haut-1)/16+1 in
  let mY = make_matrix (2*i_max) (2*j_max) (make_matrix 8 8 0) and
      mU = make_matrix i_max j_max (make_matrix 8 8 0) and
      mV = make_matrix i_max j_max (make_matrix 8 8 0) in
  for i = 0 to i_max-1 do
  for j = 0 to j_max-1 do
  mY.(2*i).(2*j) <- make_matrix 8 8 0;
  mY.(2*i).(2*j+1) <- make_matrix 8 8 0;
  mY.(2*i+1).(2*j) <- make_matrix 8 8 0;
  mY.(2*i+1).(2*j+1) <- make_matrix 8 8 0;
  mU.(i).(j) <- make_matrix 8 8 0;
  mV.(i).(j) <- make_matrix 8 8 0;
done; done;
(mY,mU,mV);;
  
(*                           LIRE UN FICHIER BMP *)

(* Programme permettant de lire les n premiers octets d'un fichier ... *)
let lire fichier n =
  let rec aux channel n = match n with
      | 0 -> close_in channel; [];
      | _ -> let x =input_byte channel in
                  x::(aux channel (n-1))
  in aux (open_in fichier) n;;

(* Lit 2 octets *)
let lire_1mot channel =
  let n1=input_byte channel in
   let n2=input_byte channel in
   256*n2+n1;;

(* Lit 4 octets *)   
let lire_2mot channel =
  let n1=input_byte channel in 
   let n2=input_byte channel in
   let n3=input_byte channel in
   let n4=input_byte channel in
  n1+256*(n2+256*(n3+256*n4));;

(* Transforme channel en matrice de matrices 8*8 *)
let bmp_to_matofmat channel larg haut =
 let (mY,mU,mV) = make_3matofmat larg haut in
for i1 = 0 to (haut-1)/8 do
 for i2 = 0 to 7 do
  for j1 = 0 to (larg-1)/8 do 
   for j2 = 0 to 3 do
 let i_2 = i2/2+4*(i1 mod 2) and j_2 = j2+4*(j1 mod 2) in
 let b1=input_byte channel in let g1=input_byte channel in let r1=input_byte channel in
 let b2=input_byte channel in let g2=input_byte channel in let r2=input_byte channel in
 let (y1,u1,v1) = rvb_to_yuv (r1,g1,b1) and (y2,u2,v2) = rvb_to_yuv (r2,g2,b2) in
  mY.(i1).(j1).(i2).(2*j2) <- y1; mY.(i1).(j1).(i2).(2*j2+1) <- y2;
  if (i2 mod 2) = 1 then (
  mU.(i1/2).(j1/2).(i_2).(j_2) <-  div_int ( mU.(i1/2).(j1/2).(i_2).(j_2) +u1+u2 ) 4;
  mV.(i1/2).(j1/2).(i_2).(j_2) <-  div_int ( mV.(i1/2).(j1/2).(i_2).(j_2) +v1+v2 ) 4; )
                    else (
  mU.(i1/2).(j1/2).(i_2).(j_2) <- u1+u2;
  mV.(i1/2).(j1/2).(i_2).(j_2) <- v1+v2; )
  done; done; done; done;
 (mY,mU,mV);;
     
(* Convertit un BMP en matrice de matrices *)
let lire_bmp str =
  print_string "Lecture du fichier en cours..."; print_newline ();
  let channel = open_in_bin str in
   if (lire_1mot channel)<>19778 then failwith "Ce n'est un fichier BMP valide";
   seek_in channel 14; let decalage = lire_2mot channel in
   let larg = lire_2mot channel in
   let haut = lire_2mot channel in
   seek_in channel 28;
   if (lire_1mot channel) <> 24 then failwith "Il faut une image 24 bits";
   seek_in channel (14+decalage);
   let (mY,mU,mV) = bmp_to_matofmat channel larg haut in
     close_in channel;
 (larg,haut,mY,mU,mV);;


(* ECRIRE UN FICHIER BMP *)

(* Ecrit un nombre sur 4 octets *)
let ecr_2mot channel n = 
  let n1=n/16777216 and p1=n mod 16777216 in
    let n2=p1/65536 and p2=p1 mod 65536 in
     let n3=p2/256 and n4=p2 mod 256 in
output_byte channel n4;
output_byte channel n3;
output_byte channel n2;
output_byte channel n1;;

(* Ecrit un nombre sur 2 octets *)
let ecr_1mot channel n =
output_byte channel (n mod 256);
output_byte channel (n/256);;

(* Se charge d'crire un en-tte BMP convenable *)  
let en_tete channel larg haut =
output_byte channel 66; output_byte channel 77;
ecr_2mot channel (larg*haut*3+56);
ecr_2mot channel 0;
ecr_2mot channel 54; ecr_2mot channel 40;
ecr_2mot channel larg; ecr_2mot channel haut;
ecr_1mot channel 1; ecr_1mot channel 24;
ecr_2mot channel 0;ecr_2mot channel 0;
ecr_2mot channel 0;ecr_2mot channel 0;
ecr_2mot channel 0;ecr_2mot channel 0;;

(* Convertit une matrice de matrices en channel *)    
let matofmat_to_bmp channel larg haut mY mU mV =
 let j_max = (larg-1)/8 and i_max=(haut-1)/8 in
  for i1=0 to i_max do
  for i2=0 to 7 do
  for j1=0 to j_max do
  for j2=0 to 7 do
  let i_2 = i2/2+4*(i1 mod 2) and j_2 = j2/2+4*(j1 mod 2) in
  let y=mY.(i1).(j1).(i2).(j2) and u=mU.(i1/2).(j1/2).(i_2).(j_2) and 
            v=mV.(i1/2).(j1/2).(i_2).(j_2) in
  let (r,g,b) = yuv_to_rvb (y,u,v) in
    output_byte channel (align b);
    output_byte channel (align g);
    output_byte channel (align r);
  done; done; done; done;;
  
(* Convertit une matrice de matrices en fichier BMP *)  
let ecrire_bmp larg haut mY mU mV nom =
 print_string "Ecriture du fichier en cours..."; print_newline ();
 let channel = open_out_bin nom in
 en_tete channel larg haut;
 matofmat_to_bmp channel larg haut mY mU mV;
 close_out channel ;;
 
(* COMPRESSION JPEG *)

(* Compresse le fichier BMP "nom" en JPEG avec un facteur de qualit Q *)
let compress nom q =
  print_string "**********Compression******************";
  let matquant = kant q 1 in
  let (larg,haut,mY,mU,mV) = lire_bmp nom in
  let i_max = (haut-1)/16 and j_max = (larg-1)/16 in
  let lstY= ref [] and lstU = ref [] and lstV = ref [] in
  print_string "
Dbut                50%                   Fin
";
  let pas = i_max*j_max/51 in let limite = ref pas in
  for i=0 to i_max do
  for j=0 to j_max do
    zig_zag ( quantiz (dct (mY.(2*i).(2*j))) matquant) lstY;
    zig_zag ( quantiz (dct (mY.(2*i).(2*j+1))) matquant) lstY;
    zig_zag ( quantiz (dct (mY.(2*i+1).(2*j))) matquant) lstY;
    zig_zag ( quantiz (dct (mY.(2*i+1).(2*j+1))) matquant) lstY;    
    zig_zag ( quantiz (dct (mU.(i).(j))) matquant) lstU;
    zig_zag ( quantiz (dct (mV.(i).(j))) matquant) lstV;
  if i*j >= !limite then
  ( flush stdout; print_string "|"; limite := !limite + pas ; )
  done;
  done;
let (hY,cY) = codage !lstY and (hU,cU) = codage !lstU and (hV,cV) = codage !lstV in
  print_string "
Compression acheve avec succs !";
  (larg,haut,hY,cY,hU,cU,hV,cV);;
  
let taille cY cU cV =
  ((list_length cY)+(list_length cU)+(list_length cV))/8 ;;

(* Dcompresse le fichier JPEG "nom" en BMP avec un facteur de qualit Q *)  
let decompress larg haut hY cY hU cU hV cV nom q=
  let lstY = ref (rev (dcode_texte hY cY)) and lstU = ref (rev (dcode_texte hU cU)) 
      and lstV = ref (rev (dcode_texte hV cV)) in
  let (mY,mU,mV) = make_3matofmat larg haut in
  print_string "**********Dcompression en cours******************";
  let matquant = kant q 1 in
  let i_max = (haut-1)/16 and j_max = (larg-1)/16 in
  print_string "
Dbut                50%                   Fin
";
  let pas = i_max*j_max/51 in let limite = ref pas in
  for i=0 to i_max do
  for j=0 to j_max do
    mY.(2*i).(2*j) <- idct (dequantiz (dezig_zag lstY) matquant) ;
    mY.(2*i).(2*j+1) <- idct (dequantiz (dezig_zag lstY) matquant);
    mY.(2*i+1).(2*j) <- idct (dequantiz (dezig_zag lstY) matquant);
    mY.(2*i+1).(2*j+1) <- idct (dequantiz (dezig_zag lstY) matquant);
    mU.(i).(j) <- idct (dequantiz (dezig_zag lstU) matquant);
    mV.(i).(j) <- idct (dequantiz (dezig_zag lstV) matquant);
    if i*j >= !limite then
  ( flush stdout; print_string "|"; limite := !limite + pas ; )
  done;
  done;
  print_string "
Dcompression acheve avec succs !
";
  ecrire_bmp larg haut mY mU mV nom;;

(* Compresse puis dcompresse l'image c:\Test.bmp *)  
let test3 nom q = 
  let (larg,haut,hY,cY,hU,cU,hV,cV) = compress nom q in
  print_newline ();
  decompress larg haut hY cY hU cU hV cV "c:\ThankYou.bmp" q;
  print_newline (); print_string "Termin avec succs !!!";
  print_newline ();
  print_string "Taille du fichier compress :";
  print_int (taille cY cU cV) ;
  print_string " octets
";;
